import abc
import sympy
import torch

import numbers
from copy import deepcopy

from ..utils import tolist


class Loss(object, metaclass=abc.ABCMeta):
    """Loss class. In Pixyz, all loss classes are required to inherit this class.

    Examples
    --------
    >>> import torch
    >>> from torch.nn import functional as F
    >>> from pixyz.distributions import Bernoulli, Normal
    >>> from pixyz.losses import StochasticReconstructionLoss, KullbackLeibler
    ...
    >>> # Set distributions
    >>> class Inference(Normal):
    ...     def __init__(self):
    ...         super().__init__(cond_var=["x"], var=["z"], name="q")
    ...         self.model_loc = torch.nn.Linear(128, 64)
    ...         self.model_scale = torch.nn.Linear(128, 64)
    ...     def forward(self, x):
    ...         return {"loc": self.model_loc(x), "scale": F.softplus(self.model_scale(x))}
    ...
    >>> class Generator(Bernoulli):
    ...     def __init__(self):
    ...         super().__init__(cond_var=["z"], var=["x"], name="p")
    ...         self.model = torch.nn.Linear(64, 128)
    ...     def forward(self, z):
    ...         return {"probs": torch.sigmoid(self.model(z))}
    ...
    >>> p = Generator()
    >>> q = Inference()
    >>> prior = Normal(loc=torch.tensor(0.), scale=torch.tensor(1.),
    ...                var=["z"], features_shape=[64], name="p_{prior}")
    ...
    >>> # Define a loss function (VAE)
    >>> reconst = StochasticReconstructionLoss(q, p)
    >>> kl = KullbackLeibler(q, prior)
    >>> loss_cls = (reconst - kl).mean()
    >>> print(loss_cls)
    mean \\left(- D_{KL} \\left[q(z|x)||p_{prior}(z) \\right] - \\mathbb{E}_{q(z|x)} \\left[\\log p(x|z) \\right] \\right)
    >>> # Evaluate this loss function
    >>> data = torch.randn(1, 128)  # Pseudo data
    >>> loss = loss_cls.eval({"x": data})
    >>> print(loss)  # doctest: +SKIP
    tensor(65.5939, grad_fn=<MeanBackward0>)

    """

    def __init__(self, p, q=None, input_var=None):
        """
        Parameters
        ----------
        p : pixyz.distributions.Distribution
            Distribution.
        q : pixyz.distributions.Distribution, defaults to None
            Distribution.
        input_var : :obj:`list` of :obj:`str`, defaults to None
            Input variables of this loss function.
            In general, users do not need to set them explicitly
            because these depend on the given distributions and each loss function.

        """
        self.p = p
        self.q = q

        if input_var is not None:
            self._input_var = input_var
        else:
            _input_var = deepcopy(p.input_var)
            if q is not None:
                _input_var += deepcopy(q.input_var)
                _input_var = sorted(set(_input_var), key=_input_var.index)
            self._input_var = _input_var

    @property
    def input_var(self):
        """list: Input variables of this distribution."""
        return self._input_var

    @property
    @abc.abstractmethod
    def _symbol(self):
        raise NotImplementedError

    @property
    def loss_text(self):
        return sympy.latex(self._symbol)

    def __str__(self):
        return self.loss_text

    def __repr__(self):
        return self.loss_text

    def __add__(self, other):
        return AddLoss(self, other)

    def __radd__(self, other):
        return AddLoss(other, self)

    def __sub__(self, other):
        return SubLoss(self, other)

    def __rsub__(self, other):
        return SubLoss(other, self)

    def __mul__(self, other):
        return MulLoss(self, other)

    def __rmul__(self, other):
        return MulLoss(other, self)

    def __truediv__(self, other):
        return DivLoss(self, other)

    def __rtruediv__(self, other):
        return DivLoss(other, self)

    def __neg__(self):
        return NegLoss(self)

    def abs(self):
        """Return an instance of :class:`pixyz.losses.losses.AbsLoss`.

        Returns
        -------
        pixyz.losses.losses.AbsLoss
            An instance of :class:`pixyz.losses.losses.AbsLoss`

        """
        return AbsLoss(self)

    def mean(self):
        """Return an instance of :class:`pixyz.losses.losses.BatchMean`.

        Returns
        -------
        pixyz.losses.losses.BatchMean
            An instance of :class:`pixyz.losses.BatchMean`

        """
        return BatchMean(self)

    def sum(self):
        """Return an instance of :class:`pixyz.losses.losses.BatchSum`.

        Returns
        -------
        pixyz.losses.losses.BatchSum
            An instance of :class:`pixyz.losses.losses.BatchSum`

        """
        return BatchSum(self)

    def expectation(self, p, input_var=None, sample_shape=torch.Size()):
        """Return an instance of :class:`pixyz.losses.Expectation`.

        Parameters
        ----------
        p : pixyz.distributions.Distribution
            Distribution for sampling.

        input_var : list
            Input variables of this loss.

        sample_shape : :obj:`list` or :obj:`NoneType`, defaults to torch.Size()
            Shape of generating samples.

        Returns
        -------
        pixyz.losses.Expectation
            An instance of :class:`pixyz.losses.Expectation`

        """
        return Expectation(p, self, input_var=input_var, sample_shape=sample_shape)

    def eval(self, x_dict={}, return_dict=False, **kwargs):
        """Evaluate the value of the loss function given inputs (:attr:`x_dict`).

        Parameters
        ----------
        x_dict : :obj:`dict`, defaults to {}
            Input variables.
        return_dict : bool, default to False.
            Whether to return samples along with the evaluated value of the loss function.

        Returns
        -------
        loss : torch.Tensor
            the evaluated value of the loss function.
        x_dict : :obj:`dict`
            All samples generated when evaluating the loss function.
            If :attr:`return_dict` is False, it is not returned.

        """

        if not(set(list(x_dict.keys())) >= set(self._input_var)):
            raise ValueError("Input keys are not valid, expected {} but got {}.".format(self._input_var,
                                                                                        list(x_dict.keys())))

        loss, x_dict = self._get_eval(x_dict, **kwargs)

        if return_dict:
            return loss, x_dict

        return loss

    @abc.abstractmethod
    def _get_eval(self, x_dict, **kwargs):
        raise NotImplementedError


class ValueLoss(Loss):
    """
    This class contains a scalar as a loss value.

    If multiplying a scalar by an arbitrary loss class, this scalar is converted to the :class:`ValueLoss`.


    Examples
    --------
    >>> loss_cls = ValueLoss(2)
    >>> print(loss_cls)
    2
    >>> loss = loss_cls.eval()
    >>> print(loss)
    2

    """
    def __init__(self, loss1):
        self.loss1 = loss1
        self._input_var = []

    def _get_eval(self, x_dict={}, **kwargs):
        return self.loss1, x_dict

    @property
    def _symbol(self):
        return self.loss1


class Parameter(Loss):
    """
    This class defines a single variable as a loss class.

    It can be used such as a coefficient parameter of a loss class.

    Examples
    --------
    >>> loss_cls = Parameter("x")
    >>> print(loss_cls)
    x
    >>> loss = loss_cls.eval({"x": 2})
    >>> print(loss)
    2

    """
    def __init__(self, input_var):
        if not isinstance(input_var, str):
            raise ValueError
        self._input_var = tolist(input_var)

    def _get_eval(self, x_dict={}, **kwargs):
        return x_dict[self._input_var[0]], x_dict

    @property
    def _symbol(self):
        return sympy.Symbol(self._input_var[0])


class LossOperator(Loss):
    def __init__(self, loss1, loss2):
        _input_var = []

        if isinstance(loss1, Loss):
            _input_var += deepcopy(loss1.input_var)
        elif isinstance(loss1, numbers.Number):
            loss1 = ValueLoss(loss1)
        elif isinstance(loss2, type(None)):
            pass
        else:
            raise ValueError("{} cannot be operated with {}.".format(type(loss1), type(loss2)))

        if isinstance(loss2, Loss):
            _input_var += deepcopy(loss2.input_var)
        elif isinstance(loss2, numbers.Number):
            loss2 = ValueLoss(loss2)
        elif isinstance(loss2, type(None)):
            pass
        else:
            raise ValueError("{} cannot be operated with {}.".format(type(loss2), type(loss1)))

        _input_var = sorted(set(_input_var), key=_input_var.index)

        self._input_var = _input_var
        self.loss1 = loss1
        self.loss2 = loss2

    def _get_eval(self, x_dict={}, **kwargs):
        if not isinstance(self.loss1, type(None)):
            loss1, x1 = self.loss1._get_eval(x_dict, **kwargs)
        else:
            loss1 = 0
            x1 = {}

        if not isinstance(self.loss2, type(None)):
            loss2, x2 = self.loss2._get_eval(x_dict, **kwargs)
        else:
            loss2 = 0
            x2 = {}

        x1.update(x2)

        return loss1, loss2, x1


class AddLoss(LossOperator):
    """
    Apply the `add` operation to the two losses.

    Examples
    --------
    >>> loss_cls_1 = ValueLoss(2)
    >>> loss_cls_2 = Parameter("x")
    >>> loss_cls = loss_cls_1 + loss_cls_2  # equals to AddLoss(loss_cls_1, loss_cls_2)
    >>> print(loss_cls)
    x + 2
    >>> loss = loss_cls.eval({"x": 3})
    >>> print(loss)
    5

    """
    @property
    def _symbol(self):
        return self.loss1._symbol + self.loss2._symbol

    def _get_eval(self, x_dict={}, **kwargs):
        loss1, loss2, x_dict = super()._get_eval(x_dict, **kwargs)
        return loss1 + loss2, x_dict


class SubLoss(LossOperator):
    """
    Apply the `sub` operation to the two losses.

    Examples
    --------
    >>> loss_cls_1 = ValueLoss(2)
    >>> loss_cls_2 = Parameter("x")
    >>> loss_cls = loss_cls_1 - loss_cls_2  # equals to SubLoss(loss_cls_1, loss_cls_2)
    >>> print(loss_cls)
    2 - x
    >>> loss = loss_cls.eval({"x": 4})
    >>> print(loss)
    -2
    >>> loss_cls = loss_cls_2 - loss_cls_1  # equals to SubLoss(loss_cls_2, loss_cls_1)
    >>> print(loss_cls)
    x - 2
    >>> loss = loss_cls.eval({"x": 4})
    >>> print(loss)
    2

    """
    @property
    def _symbol(self):
        return self.loss1._symbol - self.loss2._symbol

    def _get_eval(self, x_dict={}, **kwargs):
        loss1, loss2, x_dict = super()._get_eval(x_dict, **kwargs)
        return loss1 - loss2, x_dict


class MulLoss(LossOperator):
    """
    Apply the `mul` operation to the two losses.

    Examples
    --------
    >>> loss_cls_1 = ValueLoss(2)
    >>> loss_cls_2 = Parameter("x")
    >>> loss_cls = loss_cls_1 * loss_cls_2  # equals to MulLoss(loss_cls_1, loss_cls_2)
    >>> print(loss_cls)
    2 x
    >>> loss = loss_cls.eval({"x": 4})
    >>> print(loss)
    8

    """
    @property
    def _symbol(self):
        return self.loss1._symbol * self.loss2._symbol

    def _get_eval(self, x_dict={}, **kwargs):
        loss1, loss2, x_dict = super()._get_eval(x_dict, **kwargs)
        return loss1 * loss2, x_dict


class DivLoss(LossOperator):
    """
    Apply the `div` operation to the two losses.

    Examples
    --------
    >>> loss_cls_1 = ValueLoss(2)
    >>> loss_cls_2 = Parameter("x")
    >>> loss_cls = loss_cls_1 / loss_cls_2  # equals to DivLoss(loss_cls_1, loss_cls_2)
    >>> print(loss_cls)
    \\frac{2}{x}
    >>> loss = loss_cls.eval({"x": 4})
    >>> print(loss)
    0.5
    >>> loss_cls = loss_cls_2 / loss_cls_1  # equals to DivLoss(loss_cls_2, loss_cls_1)
    >>> print(loss_cls)
    \\frac{x}{2}
    >>> loss = loss_cls.eval({"x": 4})
    >>> print(loss)
    2.0


    """
    @property
    def _symbol(self):
        return self.loss1._symbol / self.loss2._symbol

    def _get_eval(self, x_dict={}, **kwargs):
        loss1, loss2, x_dict = super()._get_eval(x_dict, **kwargs)
        return loss1 / loss2, x_dict


class LossSelfOperator(Loss):
    def __init__(self, loss1):
        _input_var = []

        if isinstance(loss1, type(None)):
            raise ValueError

        if isinstance(loss1, Loss):
            _input_var = deepcopy(loss1.input_var)
        elif isinstance(loss1, numbers.Number):
            loss1 = ValueLoss(loss1)
        else:
            raise ValueError

        self._input_var = _input_var
        self.loss1 = loss1

    def train(self, x_dict={}, **kwargs):
        return self.loss1.train(x_dict, **kwargs)

    def test(self, x_dict={}, **kwargs):
        return self.loss1.test(x_dict, **kwargs)


class NegLoss(LossSelfOperator):
    """
    Apply the `neg` operation to the loss.

    Examples
    --------
    >>> loss_cls_1 = Parameter("x")
    >>> loss_cls = -loss_cls_1  # equals to NegLoss(loss_cls_1)
    >>> print(loss_cls)
    - x
    >>> loss = loss_cls.eval({"x": 4})
    >>> print(loss)
    -4

    """
    @property
    def _symbol(self):
        return -self.loss1._symbol

    def _get_eval(self, x_dict={}, **kwargs):
        loss, x_dict = self.loss1._get_eval(x_dict, **kwargs)
        return -loss, x_dict


class AbsLoss(LossSelfOperator):
    """
    Apply the `abs` operation to two losses.

    Examples
    --------
    >>> import torch
    >>> from pixyz.distributions import Normal
    >>> from pixyz.losses import LogProb
    >>> p = Normal(loc=torch.tensor(0.), scale=torch.tensor(1.), var=["x"],
    ...            features_shape=[10])
    >>> loss_cls = LogProb(p).abs() # equals to AbsLoss(LogProb(p))
    >>> print(loss_cls)
    |\\log p(x)|
    >>> sample_x = torch.randn(2, 10) # Psuedo data
    >>> loss = loss_cls.eval({"x": sample_x})
    >>> print(loss) # doctest: +SKIP
    tensor([12.9894, 15.5280])

    """
    @property
    def _symbol(self):
        return sympy.Symbol("|{}|".format(self.loss1.loss_text))

    def _get_eval(self, x_dict={}, **kwargs):
        loss, x_dict = self.loss1._get_eval(x_dict, **kwargs)
        return loss.abs(), x_dict


class BatchMean(LossSelfOperator):
    r"""
    Average a loss class over given batch data.

    .. math::

        \mathbb{E}_{p_{data}(x)}[\mathcal{L}(x)] \approx \frac{1}{N}\sum_{i=1}^N \mathcal{L}(x_i),

    where :math:`x_i \sim p_{data}(x)` and :math:`\mathcal{L}` is a loss function.

    Examples
    --------
    >>> import torch
    >>> from pixyz.distributions import Normal
    >>> from pixyz.losses import LogProb
    >>> p = Normal(loc=torch.tensor(0.), scale=torch.tensor(1.), var=["x"],
    ...            features_shape=[10])
    >>> loss_cls = LogProb(p).mean() # equals to BatchMean(LogProb(p))
    >>> print(loss_cls)
    mean \left(\log p(x) \right)
    >>> sample_x = torch.randn(2, 10) # Psuedo data
    >>> loss = loss_cls.eval({"x": sample_x})
    >>> print(loss) # doctest: +SKIP
    tensor(-14.5038)
    """

    @property
    def _symbol(self):
        return sympy.Symbol("mean \\left({} \\right)".format(self.loss1.loss_text))  # TODO: fix it

    def _get_eval(self, x_dict={}, **kwargs):
        loss, x_dict = self.loss1._get_eval(x_dict, **kwargs)
        return loss.mean(), x_dict


class BatchSum(LossSelfOperator):
    r"""
    Summation a loss class over given batch data.

    .. math::

        \sum_{i=1}^N \mathcal{L}(x_i),

    where :math:`x_i \sim p_{data}(x)` and :math:`\mathcal{L}` is a loss function.

    Examples
    --------
    >>> import torch
    >>> from pixyz.distributions import Normal
    >>> from pixyz.losses import LogProb
    >>> p = Normal(loc=torch.tensor(0.), scale=torch.tensor(1.), var=["x"],
    ...            features_shape=[10])
    >>> loss_cls = LogProb(p).sum() # equals to BatchSum(LogProb(p))
    >>> print(loss_cls)
    sum \left(\log p(x) \right)
    >>> sample_x = torch.randn(2, 10) # Psuedo data
    >>> loss = loss_cls.eval({"x": sample_x})
    >>> print(loss) # doctest: +SKIP
    tensor(-31.9434)
    """

    @property
    def _symbol(self):
        return sympy.Symbol("sum \\left({} \\right)".format(self.loss1.loss_text))  # TODO: fix it

    def _get_eval(self, x_dict={}, **kwargs):
        loss, x_dict = self.loss1._get_eval(x_dict, **kwargs)
        return loss.sum(), x_dict


class SetLoss(Loss):
    def __init__(self, loss):
        self.loss = loss
        self._input_var = loss.input_var

    def __getattr__(self, name):
        getattr(self.loss, name)

    def _get_eval(self, x_dict, **kwargs):
        return self.loss._get_eval(x_dict, **kwargs)

    @property
    def _symbol(self):
        return self.loss._symbol


class Expectation(Loss):
    r"""
    Expectation of a given function (Monte Carlo approximation).

    .. math::

        \mathbb{E}_{p(x)}[f(x)] \approx \frac{1}{L}\sum_{l=1}^L f(x_l),

    where :math:`x_l \sim p(x)`.

    Note that :math:`f` doesn't need to be able to sample, which is known as the law of the unconscious statistician
    (LOTUS).

    Therefore, in this class, :math:`f` is assumed to :attr:`pixyz.Loss`.

    Examples
    --------
    >>> import torch
    >>> from pixyz.distributions import Normal
    >>> from pixyz.losses import LogProb
    >>> q = Normal(loc="x", scale=torch.tensor(1.), var=["z"], cond_var=["x"],
    ...            features_shape=[10]) # q(z|x)
    >>> p = Normal(loc="z", scale=torch.tensor(1.), var=["x"], cond_var=["z"],
    ...            features_shape=[10]) # p(x|z)
    >>> loss_cls = LogProb(p).expectation(q) # equals to Expectation(q, LogProb(p))
    >>> print(loss_cls)
    \mathbb{E}_{p(z|x)} \left[\log p(x|z) \right]
    >>> sample_x = torch.randn(2, 10) # Psuedo data
    >>> loss = loss_cls.eval({"x": sample_x})
    >>> print(loss) # doctest: +SKIP
    tensor([-12.8181, -12.6062])

    """

    def __init__(self, p, f, input_var=None, sample_shape=torch.Size()):

        if input_var is None:
            input_var = list(set(p.input_var) | set(f.input_var) - set(p.var))
        self._f = f
        self.sample_shape = torch.Size(sample_shape)

        super().__init__(p, input_var=input_var)

    @property
    def _symbol(self):
        p_text = "{" + self.p.prob_text + "}"
        return sympy.Symbol("\\mathbb{{E}}_{} \\left[{} \\right]".format(p_text, self._f.loss_text))

    def _get_eval(self, x_dict={}, **kwargs):
        samples_dict = self.p.sample(x_dict, sample_shape=self.sample_shape, reparam=True, return_all=True)

        loss, loss_sample_dict = self._f.eval(samples_dict, return_dict=True, **kwargs)  # TODO: eval or _get_eval
        samples_dict.update(loss_sample_dict)

        # sum over sample_shape
        loss = loss.view(self.sample_shape.numel(), -1).mean(dim=0)

        return loss, samples_dict



import torch
import numpy as np

from .flows import Flow


class AffineCoupling(Flow):
    r"""
    Affine coupling layer

    .. math::
        :nowrap:

        \begin{eqnarray*}
        \mathbf{y}_{1:d} &=& \mathbf{x}_{1:d} \\
        \mathbf{y}_{d+1:D} &=& \mathbf{x}_{d+1:D} \odot \exp(s(\mathbf{x}_{1:d})+t(\mathbf{x}_{1:d}))
        \end{eqnarray*}

    """

    def __init__(self, in_features, mask_type="channel_wise",
                 scale_net=None, translate_net=None, scale_translate_net=None,
                 inverse_mask=False):
        super().__init__(in_features)

        # mask initializations
        if mask_type in ["checkerboard", "channel_wise"]:
            self.mask_type = mask_type
        else:
            raise ValueError

        self.inverse_mask = inverse_mask

        self.scale_net = None
        self.translate_net = None
        self.scale_translate_net = None

        if scale_net and translate_net:
            self.scale_net = scale_net
            self.translate_net = translate_net
        elif scale_translate_net:
            self.scale_translate_net = scale_translate_net
        else:
            raise ValueError

    def build_mask(self, x):
        """
        Parameters
        ----------
        x : torch.Tensor

        Returns
        -------
        mask : torch.tensor

        Examples
        --------
        >>> scale_translate_net = lambda x: (x, x)
        >>> f1 = AffineCoupling(4, mask_type="channel_wise", scale_translate_net=scale_translate_net,
        ...                     inverse_mask=False)
        >>> x1 = torch.randn([1,4,3,3])
        >>> f1.build_mask(x1)
        tensor([[[[1.]],
        <BLANKLINE>
                 [[1.]],
        <BLANKLINE>
                 [[0.]],
        <BLANKLINE>
                 [[0.]]]])
        >>> f2 = AffineCoupling(2, mask_type="checkerboard", scale_translate_net=scale_translate_net,
        ...                     inverse_mask=True)
        >>> x2 = torch.randn([1,2,5,5])
        >>> f2.build_mask(x2)
        tensor([[[[0., 1., 0., 1., 0.],
                  [1., 0., 1., 0., 1.],
                  [0., 1., 0., 1., 0.],
                  [1., 0., 1., 0., 1.],
                  [0., 1., 0., 1., 0.]]]])

        """
        if x.dim() == 4:
            [_, channels, height, width] = x.shape
            if self.mask_type == "checkerboard":
                mask = checkerboard_mask(height, width, self.inverse_mask)
                return torch.from_numpy(mask).view(1, 1, height, width).to(x.device)
            else:
                mask = channel_wise_mask(channels, self.inverse_mask)
                return torch.from_numpy(mask).view(1, channels, 1, 1).to(x.device)

        elif x.dim() == 2:
            [_, n_features] = x.shape
            if self.mask_type != "checkerboard":
                mask = channel_wise_mask(n_features, self.inverse_mask)
                return torch.from_numpy(mask).view(1, n_features).to(x.device)

        raise ValueError

    def get_parameters(self, x, y=None):
        r"""
        Parameters
        ----------
        x : torch.tensor
        y : torch.tensor

        Returns
        -------
        s : torch.tensor
        t : torch.tensor

        Examples
        --------
        >>> # In case of using scale_translate_net
        >>> scale_translate_net = lambda x: (x, x)
        >>> f1 = AffineCoupling(4, mask_type="channel_wise", scale_translate_net=scale_translate_net,
        ...                     inverse_mask=False)
        >>> x1 = torch.randn([1,4,3,3])
        >>> log_s, t = f1.get_parameters(x1)
        >>> # In case of using scale_net and translate_net
        >>> scale_net = lambda x: x
        >>> translate_net = lambda x: x
        >>> f2 = AffineCoupling(4, mask_type="channel_wise", scale_net=scale_net, translate_net=translate_net,
        ...                     inverse_mask=False)
        >>> x2 = torch.randn([1,4,3,3])
        >>> log_s, t = f2.get_parameters(x2)
        """

        if self.scale_translate_net:
            if y is None:
                log_s, t = self.scale_translate_net(x)
            else:
                log_s, t = self.scale_translate_net(x, y)
        else:
            if y is None:
                log_s = self.scale_net(x)
                t = self.translate_net(x)
            else:
                log_s = self.scale_net(x, y)
                t = self.translate_net(x, y)

        return log_s, t

    def forward(self, x, y=None, compute_jacobian=True):
        mask = self.build_mask(x)
        x_masked = mask * x
        x_inv_masked = (1 - mask) * x

        log_s, t = self.get_parameters(x_masked, y)

        log_s = log_s * (1 - mask)
        t = t * (1 - mask)

        x = x_masked + x_inv_masked * torch.exp(log_s) + t

        if compute_jacobian:
            self._logdet_jacobian = log_s.view(log_s.size(0), -1).sum(-1)

        return x

    def inverse(self, z, y=None):
        mask = self.build_mask(z)
        z_masked = mask * z
        z_inv_masked = (1 - mask) * z

        log_s, t = self.get_parameters(z_masked, y)

        log_s = log_s * (1 - mask)
        t = t * (1 - mask)

        z = z_masked + (z_inv_masked - t) * torch.exp(-log_s)

        return z

    def extra_repr(self):
        return 'in_features={}, mask_type={}, inverse_mask={}'.format(
            self.in_features, self.mask_type, self.inverse_mask
        )


def checkerboard_mask(height, width, inverse_mask=False):
    r"""
    Parameters
    ----------
    height : int
    width : int
    inverse_mask : bool

    Returns
    -------
    mask : np.array

    Examples
    --------
    >>> checkerboard_mask(5, 4, False)
    array([[1., 0., 1., 0.],
           [0., 1., 0., 1.],
           [1., 0., 1., 0.],
           [0., 1., 0., 1.],
           [1., 0., 1., 0.]], dtype=float32)
    >>> checkerboard_mask(5, 4, True)
    array([[0., 1., 0., 1.],
           [1., 0., 1., 0.],
           [0., 1., 0., 1.],
           [1., 0., 1., 0.],
           [0., 1., 0., 1.]], dtype=float32)

    """
    mask = np.arange(height).reshape(-1, 1) + np.arange(width)
    mask = np.mod((inverse_mask is False) + mask, 2)

    return mask.astype(np.float32)


def channel_wise_mask(channels, inverse_mask=False):
    r"""
    Parameters
    ----------
    channels : int
    inverse_mask : bool

    Returns
    -------
    mask : np.array

    Examples
    --------
    >>> channel_wise_mask(6, False)
    array([1., 1., 1., 0., 0., 0.], dtype=float32)
    >>> channel_wise_mask(6, True)
    array([0., 0., 0., 1., 1., 1.], dtype=float32)

    """
    mask = np.zeros(channels).astype(np.float32)
    if inverse_mask:
        mask[channels // 2:] = 1
    else:
        mask[:channels // 2] = 1

    return mask



from __future__ import print_function

from .distributions import Distribution
from ..utils import get_dict_values


class Deterministic(Distribution):
    """
    Deterministic distribution (or degeneration distribution)

    Examples
    --------
    >>> import torch
    >>> class Generator(Deterministic):
    ...     def __init__(self):
    ...         super().__init__(cond_var=["z"], var=["x"])
    ...         self.model = torch.nn.Linear(64, 512)
    ...     def forward(self, z):
    ...         return {"x": self.model(z)}
    >>> p = Generator()
    >>> print(p)
    Distribution:
      p(x|z)
    Network architecture:
      Generator(
        name=p, distribution_name=Deterministic,
        var=['x'], cond_var=['z'], input_var=['z'], features_shape=torch.Size([])
        (model): Linear(in_features=64, out_features=512, bias=True)
      )
    >>> sample = p.sample({"z": torch.randn(1, 64)})
    >>> p.log_prob().eval(sample) # log_prob is not defined.
    Traceback (most recent call last):
     ...
    NotImplementedError
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def distribution_name(self):
        return "Deterministic"

    def sample(self, x_dict={}, return_all=True, **kwargs):
        x_dict = self._check_input(x_dict)
        _x_dict = get_dict_values(x_dict, self.input_var, return_dict=True)
        output_dict = self.forward(**_x_dict)

        if set(output_dict.keys()) != set(self._var):
            raise ValueError("Output variables are not same as `var`.")

        if return_all:
            x_dict.update(output_dict)
            return x_dict

        return output_dict

    def sample_mean(self, x_dict):
        return self.sample(x_dict, return_all=False)[self._var[0]]


class DataDistribution(Distribution):
    """
    Data distribution.

    Samples from this distribution equal given inputs.

    Examples
    --------
    >>> import torch
    >>> p = DataDistribution(var=["x"])
    >>> print(p)
    Distribution:
      p_{data}(x)
    Network architecture:
      DataDistribution(
        name=p_{data}, distribution_name=Data distribution,
        var=['x'], cond_var=[], input_var=['x'], features_shape=torch.Size([])
      )
    >>> sample = p.sample({"x": torch.randn(1, 64)})
    """

    def __init__(self, var, name="p_{data}"):
        super().__init__(var=var, cond_var=[], name=name)

    @property
    def distribution_name(self):
        return "Data distribution"

    def sample(self, x_dict={}, **kwargs):
        output_dict = self._check_input(x_dict)
        return output_dict

    def sample_mean(self, x_dict):
        return self.sample(x_dict, return_all=False)[self._var[0]]

    @property
    def input_var(self):
        """
        In DataDistribution, `input_var` is same as `var`.
        """

        return self.var



import sympy
from torch import optim, nn
import torch
from .losses import Loss
from ..utils import get_dict_values, detach_dict


class AdversarialLoss(Loss):
    def __init__(self, p, q, discriminator, input_var=None,
                 optimizer=optim.Adam, optimizer_params={}):
        if p.var != q.var:
            raise ValueError("The two distribution variables must be the same.")

        if len(p.input_var) > 0:
            self.input_dist = p
        elif len(q.input_var) > 0:
            self.input_dist = q
        else:
            raise NotImplementedError

        super().__init__(p, q, input_var=input_var)

        self.loss_optimizer = optimizer
        self.loss_optimizer_params = optimizer_params
        self.d = discriminator

        params = discriminator.parameters()
        self.d_optimizer = optimizer(params, **optimizer_params)

    def _get_batch_n(self, x_dict):
        return get_dict_values(x_dict, self.input_dist.input_var[0])[0].shape[0]

    def d_loss(self, y_p, y_q, batch_n):
        """Evaluate a discriminator loss given outputs of the discriminator.

        Parameters
        ----------
        y_p : torch.Tensor
            Output of discriminator given sample from p.
        y_q : torch.Tensor
            Output of discriminator given sample from q.
        batch_n : int
            Batch size of inputs.

        Returns
        -------
        torch.Tensor

        """
        raise NotImplementedError

    def g_loss(self, y_p, y_q, batch_n):
        """Evaluate a generator loss given outputs of the discriminator.

        Parameters
        ----------
        y_p : torch.Tensor
            Output of discriminator given sample from p.
        y_q : torch.Tensor
            Output of discriminator given sample from q.
        batch_n : int
            Batch size of inputs.

        Returns
        -------
        torch.Tensor

        """
        raise NotImplementedError

    def train(self, train_x_dict, **kwargs):
        """Train the evaluation metric (discriminator).

        Parameters
        ----------
        train_x_dict : dict
            Input variables.
        **kwargs
            Arbitrary keyword arguments.

        Returns
        -------
        loss : torch.Tensor

        """
        self.d.train()

        self.d_optimizer.zero_grad()
        loss = self.eval(train_x_dict, discriminator=True)

        # backprop
        loss.backward()

        # update params
        self.d_optimizer.step()

        return loss

    def test(self, test_x_dict, **kwargs):
        """Test the evaluation metric (discriminator).

        Parameters
        ----------
        test_x_dict : dict
            Input variables.
        **kwargs
            Arbitrary keyword arguments.

        Returns
        -------
        loss : torch.Tensor

        """
        self.d.eval()

        with torch.no_grad():
            loss = self.eval(test_x_dict, discriminator=True)

        return loss


class AdversarialJensenShannon(AdversarialLoss):
    r"""
    Jensen-Shannon divergence (adversarial training).

    .. math::

        D_{JS}[p(x)||q(x)] \leq 2 \cdot D_{JS}[p(x)||q(x)] + 2 \log 2
         = \mathbb{E}_{p(x)}[\log d^*(x)] + \mathbb{E}_{q(x)}[\log (1-d^*(x))],

    where :math:`d^*(x) = \arg\max_{d} \mathbb{E}_{p(x)}[\log d(x)] + \mathbb{E}_{q(x)}[\log (1-d(x))]`.

    This class acts as a metric that evaluates a given distribution (generator).
    If you want to learn this evaluation metric itself, i.e., discriminator (critic), use the :class:`train` method.

    Examples
    --------
    >>> import torch
    >>> from pixyz.distributions import Deterministic, DataDistribution, Normal
    >>> # Generator
    >>> class Generator(Deterministic):
    ...     def __init__(self):
    ...         super(Generator, self).__init__(cond_var=["z"], var=["x"], name="p")
    ...         self.model = nn.Linear(32, 64)
    ...     def forward(self, z):
    ...         return {"x": self.model(z)}
    >>> p_g = Generator()
    >>> prior = Normal(loc=torch.tensor(0.), scale=torch.tensor(1.),
    ...                var=["z"], features_shape=[32], name="p_{prior}")
    >>> p = (p_g*prior).marginalize_var("z")
    >>> print(p)
    Distribution:
      p(x) = \int p(x|z)p_{prior}(z)dz
    Network architecture:
      Normal(
        name=p_{prior}, distribution_name=Normal,
        var=['z'], cond_var=[], input_var=[], features_shape=torch.Size([32])
        (loc): torch.Size([1, 32])
        (scale): torch.Size([1, 32])
      )
      Generator(
        name=p, distribution_name=Deterministic,
        var=['x'], cond_var=['z'], input_var=['z'], features_shape=torch.Size([])
        (model): Linear(in_features=32, out_features=64, bias=True)
      )
    >>> # Data distribution (dummy distribution)
    >>> p_data = DataDistribution(["x"])
    >>> print(p_data)
    Distribution:
      p_{data}(x)
    Network architecture:
      DataDistribution(
        name=p_{data}, distribution_name=Data distribution,
        var=['x'], cond_var=[], input_var=['x'], features_shape=torch.Size([])
      )
    >>> # Discriminator (critic)
    >>> class Discriminator(Deterministic):
    ...     def __init__(self):
    ...         super(Discriminator, self).__init__(cond_var=["x"], var=["t"], name="d")
    ...         self.model = nn.Linear(64, 1)
    ...     def forward(self, x):
    ...         return {"t": torch.sigmoid(self.model(x))}
    >>> d = Discriminator()
    >>> print(d)
    Distribution:
      d(t|x)
    Network architecture:
      Discriminator(
        name=d, distribution_name=Deterministic,
        var=['t'], cond_var=['x'], input_var=['x'], features_shape=torch.Size([])
        (model): Linear(in_features=64, out_features=1, bias=True)
      )
    >>>
    >>> # Set the loss class
    >>> loss_cls = AdversarialJensenShannon(p, p_data, discriminator=d)
    >>> print(loss_cls)
    mean(D_{JS}^{Adv} \left[p(x)||p_{data}(x) \right])
    >>>
    >>> sample_x = torch.randn(2, 64) # Psuedo data
    >>> loss = loss_cls.eval({"x": sample_x})
    >>> print(loss) # doctest: +SKIP
    tensor(1.3723, grad_fn=<AddBackward0>)
    >>> # For evaluating a discriminator loss, set the `discriminator` option to True.
    >>> loss_d = loss_cls.eval({"x": sample_x}, discriminator=True)
    >>> print(loss_d) # doctest: +SKIP
    tensor(1.4990, grad_fn=<AddBackward0>)
    >>> # When training the evaluation metric (discriminator), use the train method.
    >>> train_loss = loss_cls.train({"x": sample_x})

    References
    ----------
    [Goodfellow+ 2014] Generative Adversarial Networks
    """

    def __init__(self, p, q, discriminator, input_var=None, optimizer=optim.Adam, optimizer_params={},
                 inverse_g_loss=True):
        super().__init__(p, q, discriminator,
                         input_var=input_var,
                         optimizer=optimizer, optimizer_params=optimizer_params)

        self.bce_loss = nn.BCELoss()
        self._inverse_g_loss = inverse_g_loss

    @property
    def _symbol(self):
        return sympy.Symbol("mean(D_{{JS}}^{{Adv}} \\left[{}||{} \\right])".format(self.p.prob_text,
                                                                                   self.q.prob_text))

    def _get_eval(self, x_dict, discriminator=False, **kwargs):
        batch_n = self._get_batch_n(x_dict)

        # sample x_p from p
        x_p_dict = get_dict_values(self.p.sample(x_dict, batch_n=batch_n), self.d.input_var, True)
        # sample x_q from q
        x_q_dict = get_dict_values(self.q.sample(x_dict, batch_n=batch_n), self.d.input_var, True)
        if discriminator:
            # sample y_p from d
            y_p = get_dict_values(self.d.sample(detach_dict(x_p_dict)), self.d.var)[0]
            # sample y_q from d
            y_q = get_dict_values(self.d.sample(detach_dict(x_q_dict)), self.d.var)[0]

            return self.d_loss(y_p, y_q, batch_n), x_dict

        # sample y_p from d
        y_p_dict = self.d.sample(x_p_dict)
        # sample y_q from d
        y_q_dict = self.d.sample(x_q_dict)

        y_p = get_dict_values(y_p_dict, self.d.var)[0]
        y_q = get_dict_values(y_q_dict, self.d.var)[0]

        return self.g_loss(y_p, y_q, batch_n), x_dict

    def d_loss(self, y_p, y_q, batch_n):
        # set labels
        t_p = torch.ones(batch_n, 1).to(y_p.device)
        t_q = torch.zeros(batch_n, 1).to(y_p.device)

        return self.bce_loss(y_p, t_p) + self.bce_loss(y_q, t_q)

    def g_loss(self, y_p, y_q, batch_n):
        # set labels
        t1 = torch.ones(batch_n, 1).to(y_p.device)
        t2 = torch.zeros(batch_n, 1).to(y_p.device)

        if self._inverse_g_loss:
            y_p_loss = self.bce_loss(y_p, t2)
            y_q_loss = self.bce_loss(y_q, t1)
        else:
            y_p_loss = -self.bce_loss(y_p, t1)
            y_q_loss = -self.bce_loss(y_q, t2)

        if self.p.distribution_name == "Data distribution":
            y_p_loss = y_p_loss.detach()

        if self.q.distribution_name == "Data distribution":
            y_q_loss = y_q_loss.detach()

        return y_p_loss + y_q_loss

    def train(self, train_x_dict, **kwargs):
        return super().train(train_x_dict, **kwargs)

    def test(self, test_x_dict, **kwargs):
        return super().test(test_x_dict, **kwargs)


class AdversarialKullbackLeibler(AdversarialLoss):
    r"""
    Kullback-Leibler divergence (adversarial training).

    .. math::

        D_{KL}[p(x)||q(x)] = \mathbb{E}_{p(x)}[\log \frac{p(x)}{q(x)}]
         \approx \mathbb{E}_{p(x)}[\log \frac{d^*(x)}{1-d^*(x)}],

    where :math:`d^*(x) = \arg\max_{d} \mathbb{E}_{q(x)}[\log d(x)] + \mathbb{E}_{p(x)}[\log (1-d(x))]`.

    Note that this divergence is minimized to close :math:`p` to :math:`q`.

    Examples
    --------
    >>> import torch
    >>> from pixyz.distributions import Deterministic, DataDistribution, Normal
    >>> # Generator
    >>> class Generator(Deterministic):
    ...     def __init__(self):
    ...         super(Generator, self).__init__(cond_var=["z"], var=["x"], name="p")
    ...         self.model = nn.Linear(32, 64)
    ...     def forward(self, z):
    ...         return {"x": self.model(z)}
    >>> p_g = Generator()
    >>> prior = Normal(loc=torch.tensor(0.), scale=torch.tensor(1.),
    ...                var=["z"], features_shape=[32], name="p_{prior}")
    >>> p = (p_g*prior).marginalize_var("z")
    >>> print(p)
    Distribution:
      p(x) = \int p(x|z)p_{prior}(z)dz
    Network architecture:
      Normal(
        name=p_{prior}, distribution_name=Normal,
        var=['z'], cond_var=[], input_var=[], features_shape=torch.Size([32])
        (loc): torch.Size([1, 32])
        (scale): torch.Size([1, 32])
      )
      Generator(
        name=p, distribution_name=Deterministic,
        var=['x'], cond_var=['z'], input_var=['z'], features_shape=torch.Size([])
        (model): Linear(in_features=32, out_features=64, bias=True)
      )
    >>> # Data distribution (dummy distribution)
    >>> p_data = DataDistribution(["x"])
    >>> print(p_data)
    Distribution:
      p_{data}(x)
    Network architecture:
      DataDistribution(
        name=p_{data}, distribution_name=Data distribution,
        var=['x'], cond_var=[], input_var=['x'], features_shape=torch.Size([])
      )
    >>> # Discriminator (critic)
    >>> class Discriminator(Deterministic):
    ...     def __init__(self):
    ...         super(Discriminator, self).__init__(cond_var=["x"], var=["t"], name="d")
    ...         self.model = nn.Linear(64, 1)
    ...     def forward(self, x):
    ...         return {"t": torch.sigmoid(self.model(x))}
    >>> d = Discriminator()
    >>> print(d)
    Distribution:
      d(t|x)
    Network architecture:
      Discriminator(
        name=d, distribution_name=Deterministic,
        var=['t'], cond_var=['x'], input_var=['x'], features_shape=torch.Size([])
        (model): Linear(in_features=64, out_features=1, bias=True)
      )
    >>>
    >>> # Set the loss class
    >>> loss_cls = AdversarialKullbackLeibler(p, p_data, discriminator=d)
    >>> print(loss_cls)
    mean(D_{KL}^{Adv} \left[p(x)||p_{data}(x) \right])
    >>>
    >>> sample_x = torch.randn(2, 64) # Psuedo data
    >>> loss = loss_cls.eval({"x": sample_x})
    >>> # The evaluation value might be negative if the discriminator training is incomplete.
    >>> print(loss) # doctest: +SKIP
    tensor(-0.8377, grad_fn=<AddBackward0>)
    >>> # For evaluating a discriminator loss, set the `discriminator` option to True.
    >>> loss_d = loss_cls.eval({"x": sample_x}, discriminator=True)
    >>> print(loss_d) # doctest: +SKIP
    tensor(1.9321, grad_fn=<AddBackward0>)
    >>> # When training the evaluation metric (discriminator), use the train method.
    >>> train_loss = loss_cls.train({"x": sample_x})

    References
    ----------
    [Kim+ 2018] Disentangling by Factorising
    """

    def __init__(self, p, q, discriminator, **kwargs):
        super().__init__(p, q, discriminator, **kwargs)
        self.bce_loss = nn.BCELoss()

    @property
    def _symbol(self):
        return sympy.Symbol("mean(D_{{KL}}^{{Adv}} \\left[{}||{} \\right])".format(self.p.prob_text,
                                                                                   self.q.prob_text))

    def _get_eval(self, x_dict, discriminator=False, **kwargs):
        batch_n = self._get_batch_n(x_dict)

        # sample x_p from p
        x_p_dict = get_dict_values(self.p.sample(x_dict, batch_n=batch_n), self.d.input_var, True)

        if discriminator:
            # sample x_q from q
            x_q_dict = get_dict_values(self.q.sample(x_dict, batch_n=batch_n), self.d.input_var, True)

            # sample y_p from d
            y_p = get_dict_values(self.d.sample(detach_dict(x_p_dict)), self.d.var)[0]
            # sample y_q from d
            y_q = get_dict_values(self.d.sample(detach_dict(x_q_dict)), self.d.var)[0]

            return self.d_loss(y_p, y_q, batch_n), x_dict

        # sample y from d
        y_p = get_dict_values(self.d.sample(x_p_dict), self.d.var)[0]

        return self.g_loss(y_p, batch_n), x_dict

    def g_loss(self, y_p, batch_n):
        """Evaluate a generator loss given an output of the discriminator.

        Parameters
        ----------
        y_p : torch.Tensor
            Output of discriminator given sample from p.
        batch_n : int
            Batch size of inputs.

        Returns
        -------
        torch.Tensor

        """
        # set labels
        t_p = torch.ones(batch_n, 1).to(y_p.device)
        t_q = torch.zeros(batch_n, 1).to(y_p.device)

        y_p_loss = -self.bce_loss(y_p, t_p) + self.bce_loss(y_p, t_q)  # log (y_p) - log (1 - y_p)

        return y_p_loss

    def d_loss(self, y_p, y_q, batch_n):
        # set labels
        t_p = torch.ones(batch_n, 1).to(y_p.device)
        t_q = torch.zeros(batch_n, 1).to(y_p.device)

        return self.bce_loss(y_p, t_p) + self.bce_loss(y_q, t_q)

    def train(self, train_x_dict, **kwargs):
        return super().train(train_x_dict, **kwargs)

    def test(self, test_x_dict, **kwargs):
        return super().test(test_x_dict, **kwargs)


class AdversarialWassersteinDistance(AdversarialJensenShannon):
    r"""
    Wasserstein distance (adversarial training).

    .. math::

         W(p, q) = \sup_{||d||_{L} \leq 1} \mathbb{E}_{p(x)}[d(x)] - \mathbb{E}_{q(x)}[d(x)]

    Examples
    --------
    >>> import torch
    >>> from pixyz.distributions import Deterministic, DataDistribution, Normal
    >>> # Generator
    >>> class Generator(Deterministic):
    ...     def __init__(self):
    ...         super(Generator, self).__init__(cond_var=["z"], var=["x"], name="p")
    ...         self.model = nn.Linear(32, 64)
    ...     def forward(self, z):
    ...         return {"x": self.model(z)}
    >>> p_g = Generator()
    >>> prior = Normal(loc=torch.tensor(0.), scale=torch.tensor(1.),
    ...                var=["z"], features_shape=[32], name="p_{prior}")
    >>> p = (p_g*prior).marginalize_var("z")
    >>> print(p)
    Distribution:
      p(x) = \int p(x|z)p_{prior}(z)dz
    Network architecture:
      Normal(
        name=p_{prior}, distribution_name=Normal,
        var=['z'], cond_var=[], input_var=[], features_shape=torch.Size([32])
        (loc): torch.Size([1, 32])
        (scale): torch.Size([1, 32])
      )
      Generator(
        name=p, distribution_name=Deterministic,
        var=['x'], cond_var=['z'], input_var=['z'], features_shape=torch.Size([])
        (model): Linear(in_features=32, out_features=64, bias=True)
      )
    >>> # Data distribution (dummy distribution)
    >>> p_data = DataDistribution(["x"])
    >>> print(p_data)
    Distribution:
      p_{data}(x)
    Network architecture:
      DataDistribution(
        name=p_{data}, distribution_name=Data distribution,
        var=['x'], cond_var=[], input_var=['x'], features_shape=torch.Size([])
      )
    >>> # Discriminator (critic)
    >>> class Discriminator(Deterministic):
    ...     def __init__(self):
    ...         super(Discriminator, self).__init__(cond_var=["x"], var=["t"], name="d")
    ...         self.model = nn.Linear(64, 1)
    ...     def forward(self, x):
    ...         return {"t": self.model(x)}
    >>> d = Discriminator()
    >>> print(d)
    Distribution:
      d(t|x)
    Network architecture:
      Discriminator(
        name=d, distribution_name=Deterministic,
        var=['t'], cond_var=['x'], input_var=['x'], features_shape=torch.Size([])
        (model): Linear(in_features=64, out_features=1, bias=True)
      )
    >>>
    >>> # Set the loss class
    >>> loss_cls = AdversarialWassersteinDistance(p, p_data, discriminator=d)
    >>> print(loss_cls)
    mean(W^{Adv} \left(p(x), p_{data}(x) \right))
    >>>
    >>> sample_x = torch.randn(2, 64) # Psuedo data
    >>> loss = loss_cls.eval({"x": sample_x})
    >>> print(loss) # doctest: +SKIP
    tensor(-0.0060, grad_fn=<SubBackward0>)
    >>> # For evaluating a discriminator loss, set the `discriminator` option to True.
    >>> loss_d = loss_cls.eval({"x": sample_x}, discriminator=True)
    >>> print(loss_d) # doctest: +SKIP
    tensor(-0.3802, grad_fn=<NegBackward>)
    >>> # When training the evaluation metric (discriminator), use the train method.
    >>> train_loss = loss_cls.train({"x": sample_x})

    References
    ----------
    [Arjovsky+ 2017] Wasserstein GAN
    """

    def __init__(self, p, q, discriminator,
                 clip_value=0.01, **kwargs):
        super().__init__(p, q, discriminator, **kwargs)
        self._clip_value = clip_value

    @property
    def _symbol(self):
        return sympy.Symbol("mean(W^{{Adv}} \\left({}, {} \\right))".format(self.p.prob_text, self.q.prob_text))

    def d_loss(self, y_p, y_q, *args, **kwargs):
        return - (torch.mean(y_p) - torch.mean(y_q))

    def g_loss(self, y_p, y_q, *args, **kwargs):
        if self.p.distribution_name == "Data distribution":
            y_p = y_p.detach()

        if self.q.distribution_name == "Data distribution":
            y_q = y_q.detach()
        return torch.mean(y_p) - torch.mean(y_q)

    def train(self, train_x_dict, **kwargs):
        loss = super().train(train_x_dict, **kwargs)

        # Clip weights of discriminator
        for params in self.d.parameters():
            params.data.clamp_(-self._clip_value, self._clip_value)

        return loss

    def test(self, test_x_dict, **kwargs):
        return super().test(test_x_dict, **kwargs)



import torch
from torch import nn
import numpy as np

from .flows import Flow
from ..utils import epsilon


class BatchNorm1d(Flow):
    """
    A batch normalization with the inverse transformation.

    Notes
    -----
    This is implemented with reference to the following code.
    https://github.com/ikostrikov/pytorch-flows/blob/master/flows.py#L205

    Examples
    --------
    >>> x = torch.randn(20, 100)
    >>> f = BatchNorm1d(100)
    >>> # transformation
    >>> z = f(x)
    >>> # reconstruction
    >>> _x = f.inverse(f(x))
    >>> # check this reconstruction
    >>> diff = torch.sum(torch.abs(_x-x)).data
    >>> diff < 0.1
    tensor(1, dtype=torch.uint8)
    """
    def __init__(self, in_features, momentum=0.0):
        super().__init__(in_features)

        self.log_gamma = nn.Parameter(torch.zeros(in_features))
        self.beta = nn.Parameter(torch.zeros(in_features))
        self.momentum = momentum

        self.register_buffer('running_mean', torch.zeros(in_features))
        self.register_buffer('running_var', torch.ones(in_features))

    def forward(self, x, y=None, compute_jacobian=True):
        if self.training:
            self.batch_mean = x.mean(0)
            self.batch_var = (x - self.batch_mean).pow(2).mean(0) + epsilon()

            self.running_mean = self.running_mean * self.momentum
            self.running_var = self.running_var * self.momentum

            self.running_mean = self.running_mean + (self.batch_mean.data * (1 - self.momentum))
            self.running_var = self.running_var + (self.batch_var.data * (1 - self.momentum))

            mean = self.batch_mean
            var = self.batch_var

        else:
            mean = self.running_mean
            var = self.running_var

        x_hat = (x - mean) / var.sqrt()
        z = torch.exp(self.log_gamma) * x_hat + self.beta

        if compute_jacobian:
            self._logdet_jacobian = (self.log_gamma - 0.5 * torch.log(var)).sum(-1)

        return z

    def inverse(self, z, y=None):
        if self.training:
            mean = self.batch_mean
            var = self.batch_var
        else:
            mean = self.running_mean
            var = self.running_var

        x_hat = (z - self.beta) / torch.exp(self.log_gamma)

        x = x_hat * var.sqrt() + mean

        return x


class BatchNorm2d(BatchNorm1d):
    """
    A batch normalization with the inverse transformation.

    Notes
    -----
    This is implemented with reference to the following code.
    https://github.com/ikostrikov/pytorch-flows/blob/master/flows.py#L205

    Examples
    --------
    >>> x = torch.randn(20, 100, 35, 45)
    >>> f = BatchNorm2d(100)
    >>> # transformation
    >>> z = f(x)
    >>> # reconstruction
    >>> _x = f.inverse(f(x))
    >>> # check this reconstruction
    >>> diff = torch.sum(torch.abs(_x-x)).data
    >>> diff < 0.1
    tensor(1, dtype=torch.uint8)
    """
    def __init__(self, in_features, momentum=0.0):
        super().__init__(in_features, momentum)
        self.log_gamma = nn.Parameter(self._unsqueeze(self.log_gamma.data))
        self.beta = nn.Parameter(self._unsqueeze(self.beta.data))

        self.register_buffer('running_mean', self._unsqueeze(self.running_mean))
        self.register_buffer('running_var', self._unsqueeze(self.running_var))

    def _unsqueeze(self, x):
        return x.unsqueeze(1).unsqueeze(2)


class ActNorm2d(Flow):
    """
    Activation Normalization
    Initialize the bias and scale with a given minibatch,
    so that the output per-channel have zero mean and unit variance for that.
    After initialization, `bias` and `logs` will be trained as parameters.

    Notes
    -----
    This is implemented with reference to the following code.
    https://github.com/chaiyujin/glow-pytorch/blob/master/glow/modules.py
    """

    def __init__(self, in_features, scale=1.):
        super().__init__(in_features)
        # register mean and scale
        size = [1, in_features, 1, 1]
        self.register_parameter("bias", nn.Parameter(torch.zeros(*size)))
        self.register_parameter("logs", nn.Parameter(torch.zeros(*size)))
        self.scale = float(scale)
        self.inited = False

    def initialize_parameters(self, x):
        if not self.training:
            return
        assert x.device == self.bias.device
        with torch.no_grad():
            bias = torch.mean(x.clone(), dim=[0, 2, 3], keepdim=True) * -1.0
            vars = torch.mean((x.clone() + bias) ** 2, dim=[0, 2, 3], keepdim=True)
            logs = torch.log(self.scale / (torch.sqrt(vars) + epsilon()))
            self.bias.data.copy_(bias.data)
            self.logs.data.copy_(logs.data)
            self.inited = True

    def _center(self, x, inverse=False):
        if not inverse:
            return x + self.bias
        else:
            return x - self.bias

    def _scale(self, x, compute_jacobian=True, inverse=False):
        logs = self.logs
        if not inverse:
            x = x * torch.exp(logs)
        else:
            x = x * torch.exp(-logs)
        if compute_jacobian:
            """
            logs is log_std of `mean of channels`
            so we need to multiply pixels
            """
            pixels = np.prod(x.size()[2:])
            logdet_jacobian = torch.sum(logs) * pixels

            return x, logdet_jacobian

        return x, None

    def forward(self, x, y=None, compute_jacobian=True):
        if not self.inited:
            self.initialize_parameters(x)

        # center and scale
        x = self._center(x, inverse=False)
        x, logdet_jacobian = self._scale(x, compute_jacobian, inverse=False)
        if compute_jacobian:
            self._logdet_jacobian = logdet_jacobian

        return x

    def inverse(self, x, y=None):
        if not self.inited:
            self.initialize_parameters(x)

        # scale and center
        x, _ = self._scale(x, compute_jacobian=False, inverse=True)
        x = self._center(x, inverse=True)
        return x



import io
import os
import re
from setuptools import setup, find_packages


def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='pixyz',
    version=find_version("pixyz", "__init__.py"),
    packages=find_packages(),
    url='https://github.com/masa-su/pixyz',
    author='masa-su',
    author_email='masa@weblab.t.u-tokyo.ac.jp',
    description='Deep generative modeling library',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "torch>=1.0",
        "torchvision",
        "tqdm",
        "scipy",
        "sympy>=1.4",
        "ipython",
        "tensorboardX",
    ],
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
        "Operating System :: OS Independent",
    ],
)



import torch
import torch.nn.functional as F
import numpy as np

from .flows import Flow
from ..utils import sum_samples


class Squeeze(Flow):
    """
    Squeeze operation.

    c * s * s -> 4c * s/2 * s/2

    Examples
    --------
    >>> import torch
    >>> a = torch.tensor([i+1 for i in range(16)]).view(1,1,4,4)
    >>> print(a)
    tensor([[[[ 1,  2,  3,  4],
              [ 5,  6,  7,  8],
              [ 9, 10, 11, 12],
              [13, 14, 15, 16]]]])
    >>> f = Squeeze()
    >>> print(f(a))
    tensor([[[[ 1,  3],
              [ 9, 11]],
    <BLANKLINE>
             [[ 2,  4],
              [10, 12]],
    <BLANKLINE>
             [[ 5,  7],
              [13, 15]],
    <BLANKLINE>
             [[ 6,  8],
              [14, 16]]]])

    >>> print(f.inverse(f(a)))
    tensor([[[[ 1,  2,  3,  4],
              [ 5,  6,  7,  8],
              [ 9, 10, 11, 12],
              [13, 14, 15, 16]]]])

    """

    def __init__(self):
        super().__init__(None)
        self._logdet_jacobian = 0

    def forward(self, x, y=None, compute_jacobian=True):
        [_, channels, height, width] = x.shape

        if height % 2 != 0 or width % 2 != 0:
            raise ValueError

        x = x.permute(0, 2, 3, 1)

        x = x.view(-1, height // 2, 2, width // 2, 2, channels)
        x = x.permute(0, 1, 3, 5, 2, 4)
        x = x.contiguous().view(-1, height // 2, width // 2, channels * 4)

        z = x.permute(0, 3, 1, 2)

        return z

    def inverse(self, z, y=None):
        [_, channels, height, width] = z.shape

        if channels % 4 != 0:
            raise ValueError

        z = z.permute(0, 2, 3, 1)

        z = z.view(-1, height, width, channels // 4, 2, 2)
        z = z.permute(0, 1, 4, 2, 5, 3)
        z = z.contiguous().view(-1, 2 * height, 2 * width, channels // 4)

        x = z.permute(0, 3, 1, 2)

        return x


class Unsqueeze(Squeeze):
    """
    Unsqueeze operation.

    c * s * s -> c/4 * 2s * 2s

    Examples
    --------
    >>> import torch
    >>> a = torch.tensor([i+1 for i in range(16)]).view(1,4,2,2)
    >>> print(a)
    tensor([[[[ 1,  2],
              [ 3,  4]],
    <BLANKLINE>
             [[ 5,  6],
              [ 7,  8]],
    <BLANKLINE>
             [[ 9, 10],
              [11, 12]],
    <BLANKLINE>
             [[13, 14],
              [15, 16]]]])
    >>> f = Unsqueeze()
    >>> print(f(a))
    tensor([[[[ 1,  5,  2,  6],
              [ 9, 13, 10, 14],
              [ 3,  7,  4,  8],
              [11, 15, 12, 16]]]])
    >>> print(f.inverse(f(a)))
    tensor([[[[ 1,  2],
              [ 3,  4]],
    <BLANKLINE>
             [[ 5,  6],
              [ 7,  8]],
    <BLANKLINE>
             [[ 9, 10],
              [11, 12]],
    <BLANKLINE>
             [[13, 14],
              [15, 16]]]])

    """

    def forward(self, x, y=None, compute_jacobian=True):
        return super().inverse(x)

    def inverse(self, z, y=None):
        return super().forward(z)


class Permutation(Flow):
    """
    Examples
    --------
    >>> import torch
    >>> a = torch.tensor([i+1 for i in range(16)]).view(1,4,2,2)
    >>> print(a)
    tensor([[[[ 1,  2],
              [ 3,  4]],
    <BLANKLINE>
             [[ 5,  6],
              [ 7,  8]],
    <BLANKLINE>
             [[ 9, 10],
              [11, 12]],
    <BLANKLINE>
             [[13, 14],
              [15, 16]]]])
    >>> perm = [0,3,1,2]
    >>> f = Permutation(perm)
    >>> f(a)
    tensor([[[[ 1,  2],
              [ 3,  4]],
    <BLANKLINE>
             [[13, 14],
              [15, 16]],
    <BLANKLINE>
             [[ 5,  6],
              [ 7,  8]],
    <BLANKLINE>
             [[ 9, 10],
              [11, 12]]]])
    >>> f.inverse(f(a))
    tensor([[[[ 1,  2],
              [ 3,  4]],
    <BLANKLINE>
             [[ 5,  6],
              [ 7,  8]],
    <BLANKLINE>
             [[ 9, 10],
              [11, 12]],
    <BLANKLINE>
             [[13, 14],
              [15, 16]]]])

    """

    def __init__(self, permute_indices):
        super().__init__(len(permute_indices))
        self.permute_indices = permute_indices
        self.inv_permute_indices = np.argsort(self.permute_indices)
        self._logdet_jacobian = 0

    def forward(self, x, y=None, compute_jacobian=True):
        if x.dim() == 2:
            return x[:, self.permute_indices]
        elif x.dim() == 4:
            return x[:, self.permute_indices, :, :]
        raise ValueError

    def inverse(self, z, y=None):
        if z.dim() == 2:
            return z[:, self.inv_permute_indices]
        elif z.dim() == 4:
            return z[:, self.inv_permute_indices, :, :]
        raise ValueError


class Shuffle(Permutation):
    def __init__(self, in_features):
        permute_indices = np.random.permutation(in_features)
        super().__init__(permute_indices)


class Reverse(Permutation):
    def __init__(self, in_features):
        permute_indices = np.array(np.arange(0, in_features)[::-1])
        super().__init__(permute_indices)


class Flatten(Flow):
    def __init__(self, in_size=None):
        super().__init__(None)
        self.in_size = in_size
        self._logdet_jacobian = 0

    def forward(self, x, y=None, compute_jacobian=True):
        self.in_size = x.shape[1:]
        return x.view(x.size(0), -1)

    def inverse(self, z, y=None):
        if self.in_size is None:
            raise ValueError
        return z.view(z.size(0), self.in_size[0], self.in_size[1], self.in_size[2])


class Preprocess(Flow):
    def __init__(self):
        super().__init__(None)
        self.register_buffer('data_constraint', torch.tensor([0.05], dtype=torch.float32))

    @staticmethod
    def logit(x):
        return x.log() - (1. - x).log()

    def forward(self, x, y=None, compute_jacobian=True):
        # 1. transform the domain of x from [0, 1] to [0, 255]
        x = x * 255

        # 2-1. add noise to pixels to dequantize them and transform its domain ([0, 255]->[0, 1]).
        x = (x + torch.rand_like(x)) / 256.

        # 2-2. transform pixel values with logit to be unconstrained ([0, 1]->(0, 1)).
        x = (1 + (2 * x - 1) * (1 - self.data_constraint)) / 2.

        # 2-3. apply the logit function ((0, 1)->(-inf, inf)).
        z = self.logit(x)

        if compute_jacobian:
            # log-det Jacobian of transformation (2)
            logdet_jacobian = F.softplus(z) + F.softplus(-z) \
                - F.softplus(self.data_constraint.log() - (1. - self.data_constraint).log())
            logdet_jacobian = sum_samples(logdet_jacobian)

            # log-det Jacobian of transformation (1)
            logdet_jacobian = logdet_jacobian - np.log(256.) * z[0].numel()

            self._logdet_jacobian = logdet_jacobian

        return z

    def inverse(self, z, y=None):
        # transform the domain of z from (-inf, inf) to (0, 1).
        return torch.sigmoid(z)



import torch
from ..distributions import Distribution
from ..utils import get_dict_values


class TransformedDistribution(Distribution):
    r"""
    Convert flow transformations to distributions.

    .. math::

        p(z=f_{flow}(x)),

    where :math:`x \sim p_{prior}(x)`.


    Once initializing, it can be handled as a distribution module.

    """

    def __init__(self, prior, flow, var, name="p"):
        if flow.in_features:
            features_shape = [flow.in_features]
        else:
            features_shape = torch.Size()

        super().__init__(var=var, cond_var=prior.cond_var, name=name, features_shape=features_shape)
        self.prior = prior
        self.flow = flow  # FlowList

        self._flow_input_var = prior.var

    @property
    def distribution_name(self):
        return "TransformedDistribution"

    @property
    def flow_input_var(self):
        """list: Input variables of the flow module."""
        return self._flow_input_var

    @property
    def prob_factorized_text(self):
        flow_text = "{}=f_{{flow}}({})".format(self.var[0], self.flow_input_var[0])
        prob_text = "{}({})".format(self._name, flow_text)

        return prob_text

    @property
    def logdet_jacobian(self):
        """
        Get log-determinant Jacobian.

        Before calling this, you should run :attr:`forward` or :attr:`update_jacobian` methods to calculate and
        store log-determinant Jacobian.

        """
        return self.flow.logdet_jacobian

    def sample(self, x_dict={}, batch_n=None, sample_shape=torch.Size(), return_all=True, reparam=False,
               compute_jacobian=True):
        # sample from the prior
        sample_dict = self.prior.sample(x_dict, batch_n=batch_n, sample_shape=sample_shape, return_all=return_all)

        # flow transformation
        _x = get_dict_values(sample_dict, self.flow_input_var)[0]
        z = self.forward(_x, compute_jacobian=compute_jacobian)
        output_dict = {self.var[0]: z}

        if return_all:
            sample_dict.update(output_dict)
            return sample_dict

        return output_dict

    def get_log_prob(self, x_dict, sum_features=True, feature_dims=None, compute_jacobian=False):
        # prior
        log_prob_prior = self.prior.get_log_prob(x_dict, sum_features=sum_features, feature_dims=feature_dims)

        # flow
        if compute_jacobian:
            self.sample(x_dict, return_all=False, compute_jacobian=True)

        return log_prob_prior - self.logdet_jacobian

    def forward(self, x, y=None, compute_jacobian=True):
        """
        Forward propagation of flow layers.

        Parameters
        ----------
        x : torch.Tensor
            Input data.
        y : torch.Tensor, defaults to None
            Data for conditioning.
        compute_jacobian : bool, defaults to True
            Whether to calculate and store log-determinant Jacobian.
            If true, calculated Jacobian values are stored in :attr:`logdet_jacobian`.

        Returns
        -------
        z : torch.Tensor

        """
        return self.flow.forward(x=x, y=y, compute_jacobian=compute_jacobian)

    def inverse(self, z, y=None):
        """
        Backward (inverse) propagation of flow layers.
        In this method, log-determinant Jacobian is not calculated.

        Parameters
        ----------
        z : torch.Tensor
            Input data.
        y : torch.Tensor, defaults to None
            Data for conditioning.

        Returns
        -------
        x : torch.Tensor

        """
        return self.flow.inverse(z=z, y=y)


class InverseTransformedDistribution(Distribution):
    r"""
    Convert inverse flow transformations to distributions.

    .. math::

        p(x=f^{-1}_{flow}(z)),

    where :math:`z \sim p_{prior}(z)`.

    Once initializing, it can be handled as a distribution module.

    Moreover, this distribution can take a conditional variable.

    .. math::

        p(x=f^{-1}_{flow}(z, y)),

    where :math:`z \sim p_{prior}(z)` and :math:`y` is given.

    """

    def __init__(self, prior, flow, var, cond_var=[], name="p"):
        if flow.in_features:
            features_shape = [flow.in_features]
        else:
            features_shape = torch.Size()

        super().__init__(var, cond_var=cond_var, name=name, features_shape=features_shape)
        self.prior = prior
        self.flow = flow  # FlowList

        self._flow_output_var = prior.var

    @property
    def distribution_name(self):
        return "InverseTransformedDistribution"

    @property
    def flow_output_var(self):
        return self._flow_output_var

    @property
    def prob_factorized_text(self):
        var_text = ','.join(self.flow_output_var + self.cond_var)
        flow_text = "{}=f^{{-1}}_{{flow}}({})".format(self.var[0], var_text)
        prob_text = "{}({})".format(self._name, flow_text)

        return prob_text

    @property
    def logdet_jacobian(self):
        """
        Get log-determinant Jacobian.

        Before calling this, you should run :attr:`forward` or :attr:`update_jacobian` methods to calculate and
        store log-determinant Jacobian.

        """
        return self.flow.logdet_jacobian

    def sample(self, x_dict={}, batch_n=None, sample_shape=torch.Size(), return_all=True, reparam=False):
        # sample from the prior
        sample_dict = self.prior.sample(x_dict, batch_n=batch_n, sample_shape=sample_shape, return_all=return_all)

        # inverse flow transformation
        _z = get_dict_values(sample_dict, self.flow_output_var)
        _y = get_dict_values(sample_dict, self.cond_var)

        if len(_y) == 0:
            x = self.inverse(_z[0])
        else:
            x = self.inverse(_z[0], y=_y[0])

        output_dict = {self.var[0]: x}

        if return_all:
            sample_dict.update(output_dict)
            return sample_dict

        return output_dict

    def inference(self, x_dict, return_all=True, compute_jacobian=False):
        # flow transformation
        _x = get_dict_values(x_dict, self.var)
        _y = get_dict_values(x_dict, self.cond_var)

        if len(_y) == 0:
            z = self.forward(_x[0], compute_jacobian=compute_jacobian)
        else:
            z = self.forward(_x[0], y=_y[0], compute_jacobian=compute_jacobian)

        output_dict = {self.flow_output_var[0]: z}

        if return_all:
            output_dict.update(x_dict)

        return output_dict

    def get_log_prob(self, x_dict, sum_features=True, feature_dims=None):
        # flow
        output_dict = self.inference(x_dict, return_all=True, compute_jacobian=True)

        # prior
        log_prob_prior = self.prior.get_log_prob(output_dict, sum_features=sum_features, feature_dims=feature_dims)

        return log_prob_prior + self.logdet_jacobian

    def forward(self, x, y=None, compute_jacobian=True):
        """
        Forward propagation of flow layers.

        Parameters
        ----------
        x : torch.Tensor
            Input data.
        y : torch.Tensor, defaults to None
            Data for conditioning.
        compute_jacobian : bool, defaults to True
            Whether to calculate and store log-determinant Jacobian.
            If true, calculated Jacobian values are stored in :attr:`logdet_jacobian`.

        Returns
        -------
        z : torch.Tensor

        """
        return self.flow.forward(x=x, y=y, compute_jacobian=compute_jacobian)

    def inverse(self, z, y=None):
        """
        Backward (inverse) propagation of flow layers.
        In this method, log-determinant Jacobian is not calculated.

        Parameters
        ----------
        z : torch.Tensor
            Input data.
        y : torch.Tensor, defaults to None
            Data for conditioning.

        Returns
        -------
        x : torch.Tensor

        """
        return self.flow.inverse(z=z, y=y)



import torch
from torch import nn
from torch.nn import functional as F

import numpy as np
import scipy as sp

from .flows import Flow


class ChannelConv(Flow):
    """
    Invertible 1 × 1 convolution.

    Notes
    -----
    This is implemented with reference to the following code.
    https://github.com/chaiyujin/glow-pytorch/blob/master/glow/modules.py
    """

    def __init__(self, in_channels, decomposed=False):
        super().__init__(in_channels)
        w_shape = [in_channels, in_channels]
        w_init = np.linalg.qr(np.random.randn(*w_shape))[0].astype(np.float32)
        if not decomposed:
            # Sample a random orthogonal matrix:
            self.register_parameter("weight", nn.Parameter(torch.Tensor(w_init)))
        else:
            # LU decomposition
            np_p, np_l, np_u = sp.linalg.lu(w_init)
            np_s = np.diag(np_u)
            np_sign_s = np.sign(np_s)
            np_log_s = np.log(np.abs(np_s))
            np_u = np.triu(np_u, k=1)
            l_mask = np.tril(np.ones(w_shape, dtype=np.float32), -1)
            eye = np.eye(*w_shape, dtype=np.float32)

            self.register_buffer('p', torch.Tensor(np_p.astype(np.float32)))
            self.register_buffer('sign_s', torch.Tensor(np_sign_s.astype(np.float32)))

            self.l = nn.Parameter(torch.Tensor(np_l.astype(np.float32)))
            self.log_s = nn.Parameter(torch.Tensor(np_log_s.astype(np.float32)))
            self.u = nn.Parameter(torch.Tensor(np_u.astype(np.float32)))
            self.l_mask = torch.Tensor(l_mask)
            self.eye = torch.Tensor(eye)
        self.w_shape = w_shape
        self.decomposed = decomposed

    def get_parameters(self, x, inverse):
        w_shape = self.w_shape
        pixels = np.prod(x.size()[2:])
        device = x.device

        if not self.decomposed:
            logdet_jacobian = torch.slogdet(self.weight.cpu())[1].to(device) * pixels
            if not inverse:
                weight = self.weight.view(w_shape[0], w_shape[1], 1, 1)
            else:
                weight = torch.inverse(self.weight.double()).float().view(w_shape[0], w_shape[1], 1, 1)
            return weight, logdet_jacobian
        else:
            self.p = self.p.to(device)
            self.sign_s = self.sign_s.to(device)
            self.l_mask = self.l_mask.to(device)
            self.eye = self.eye.to(device)
            l = self.l * self.l_mask + self.eye
            u = self.u * self.l_mask.transpose(0, 1).contiguous() + torch.diag(self.sign_s * torch.exp(self.log_s))
            logdet_jacobian = torch.sum(self.log_s) * pixels
            if not inverse:
                w = torch.matmul(self.p, torch.matmul(l, u))
            else:
                l = torch.inverse(l.double()).float()
                u = torch.inverse(u.double()).float()
                w = torch.matmul(u, torch.matmul(l, self.p.inverse()))
            return w.view(w_shape[0], w_shape[1], 1, 1), logdet_jacobian

    def forward(self, x, y=None, compute_jacobian=True):
        weight, logdet_jacobian = self.get_parameters(x, inverse=False)
        z = F.conv2d(x, weight)

        if compute_jacobian:
            self._logdet_jacobian = logdet_jacobian

        return z

    def inverse(self, x, y=None):
        weight, _ = self.get_parameters(x, inverse=True)
        z = F.conv2d(x, weight)
        return z
